module jp.ku6.pmdloader.math;

import std.math;
import std.stdio;

/++
 + 行列(4 * 4)
 +
 + 要素名と位置の関係は次の通り
 + _11, _12, _13, _14
 + _21, _22, _23, _24
 + _31, _32, _33, _34
 + _41, _42, _43, _44
 +
 + 実際のメモリ上での並び順は下記のようになっており、Shaderに直接渡せる
 + _11, _21, _31, _41, _12, _22, _32, _42, _13, _23, _33, _43, _14, _24, _34, _44
 +/
struct Matrix
{
	/// 各要素
	float _11, _21, _31, _41, _12, _22, _32, _42, _13, _23, _33, _43, _14, _24, _34, _44;

	/++
	 + コンストラクタ
	 +
	 + 内部のメモリ配列を気にせず、行列の並び通りに設定できる
	 +
	 + Matrix
	 + (
	 +     _11, _12, _13, _14, 
	 +     _21, _22, _23, _24, 
	 +     _31, _32, _33, _34, 
	 +     _41, _42, _43, _44
	 + );
	 +/
	this(float e11, float e12, float e13, float e14, float e21, float e22, float e23, float e24, float e31, float e32, float e33, float e34, float e41, float e42, float e43, float e44)
	{
		_11 = e11;
		_12 = e12;
		_13 = e13;
		_14 = e14;
		_21 = e21;
		_22 = e22;
		_23 = e23;
		_24 = e24;
		_31 = e31;
		_32 = e32;
		_33 = e33;
		_34 = e34;
		_41 = e41;
		_42 = e42;
		_43 = e43;
		_44 = e44;
	}

	/++
	 + 転置行列を返す
	 +/
	const
	Matrix transposed()
	{
		return Matrix(_11, _21, _31, _41, _12, _22, _32, _42, _13, _23, _33, _43, _14, _24, _34, _44);
	}

	/++
	 + 行列との積
	 +/
	const
	Matrix opMul(ref const(Matrix) matrix)
	{
		return Matrix
		(
			_11 * matrix._11 + _12 * matrix._21 + _13 * matrix._31 + _14 * matrix._41,
			_11 * matrix._12 + _12 * matrix._22 + _13 * matrix._32 + _14 * matrix._42,
			_11 * matrix._13 + _12 * matrix._23 + _13 * matrix._33 + _14 * matrix._43,
			_11 * matrix._14 + _12 * matrix._24 + _13 * matrix._34 + _14 * matrix._44,

			_21 * matrix._11 + _22 * matrix._21 + _23 * matrix._31 + _24 * matrix._41,
			_21 * matrix._12 + _22 * matrix._22 + _23 * matrix._32 + _24 * matrix._42,
			_21 * matrix._13 + _22 * matrix._23 + _23 * matrix._33 + _24 * matrix._43,
			_21 * matrix._14 + _22 * matrix._24 + _23 * matrix._34 + _24 * matrix._44,

			_31 * matrix._11 + _32 * matrix._21 + _33 * matrix._31 + _34 * matrix._41,
			_31 * matrix._12 + _32 * matrix._22 + _33 * matrix._32 + _34 * matrix._42,
			_31 * matrix._13 + _32 * matrix._23 + _33 * matrix._33 + _34 * matrix._43,
			_31 * matrix._14 + _32 * matrix._24 + _33 * matrix._34 + _34 * matrix._44,

			_41 * matrix._11 + _42 * matrix._21 + _43 * matrix._31 + _44 * matrix._41,
			_41 * matrix._12 + _42 * matrix._22 + _43 * matrix._32 + _44 * matrix._42,
			_41 * matrix._13 + _42 * matrix._23 + _43 * matrix._33 + _44 * matrix._43,
			_41 * matrix._14 + _42 * matrix._24 + _43 * matrix._34 + _44 * matrix._44
		);
	}

}

/++
 + 3次元ベクトル
 +/
struct Vector3
{
	/// X軸成分
	float x;

	/// Y軸成分
	float y;

	/// Z軸成分
	float z;

	/++
	 + 反転
	 +/
	const
	Vector3 opNeg()
	{
		return Vector3(-x, -y, -z);
	}

	/++
	 + 加算
	 +/
	const
	Vector3 opAdd(ref const(Vector3) source)
	{
		return Vector3(x + source.x, y + source.y, z + source.z); 
	}

	/++
	 + 減算
	 +/
	const
	Vector3 opSub(ref const(Vector3) source)
	{
		return Vector3(x - source.x, y - source.y, z - source.z); 
	}

	/++
	 + 内積
	 +/
	const
	float opMul(ref const(Vector3) source)
	{
		return x * source.x + y * source.y + z * source.z;
	}

	/++
	 + 行列との積
	 +/
	const
	Vector3 opMul(ref const(Matrix) matrix)
	{
		return Vector3
		(
			x * matrix._11 + y * matrix._21 + z * matrix._31 + matrix._41,
			x * matrix._12 + y * matrix._22 + z * matrix._32 + matrix._42,
			x * matrix._13 + y * matrix._23 + z * matrix._33 + matrix._43
		);
	}

	/++
	 + 外積
	 +/
	const
	Vector3 opDiv(ref const(Vector3) source)
	{
		return Vector3(y * source.z - source.y * z, z * source.x - source.z * x, x * source.y - source.x * y);
	}

	/++
	 + 正規化されたベクトルを返す
	 +/
	const
	Vector3 normal()
	{
		float l = sqrt(x * x + y * y + z * z); 
		return Vector3(x / l, y / l, z / l);
	}
}

/++
 + ビュー変換用の行列を作成
 +
 + Params:
 +  camera = ワールド座標上のカメラの位置
 +  cameraUp = カメラの上方向ベクトル
 +  viewTarget = ワールド座標上の視点の位置
 +
 + Returns:
 +  ビュー変換行列
 +/
Matrix createViewMatrix(ref const(Vector3) camera, ref const(Vector3) cameraUp, ref const(Vector3) viewTarget)
{
	// 視点をカメラを中心とした座標系に移動すると、視線方向のベクトルとなる
	Vector3 Z = viewTarget - camera;

	// それを正規化する
	Z = Z.normal;

	// 視線とカメラの上方向の外積がX軸の方向となる
	Vector3 X = cameraUp / Z;

	// こちらも正規化する
	X = X.normal;

	// X軸とZ軸の単位ベクトルからY軸の単位ベクトルを計算する
	Vector3 Y = Z / X;

	// 求めた値を返す
	return Matrix
	(
		X.x, Y.x, Z.x, 0,
		X.y, Y.y, Z.y, 0,
		X.z, Y.z, Z.z, 0,
		-camera * X, -camera * Y, -camera * Z, 1
	);
}

/++
 + 射影変換用の行列を作成
 +
 + Params:
 +  vertical = 垂直視野角
 +  near = 前方クリップ位置
 +  far = 後方クリップ位置
 +  width = 画面横幅
 +  height = 画面高さ
 +
 + Returns:
 +  射影変換行列
 +/
Matrix createProjectionMatrix(float flovY, float near, float far, float width, float height)
{
	float h = 1.0f / tan(flovY * 0.5);
	float w = height / width * h;
	float Q = far / (far - near);

	return Matrix
	(
		w, 0, 0, 0,
		0, h, 0, 0,
		0, 0, Q, 1,
		0, 0, -Q * near, 1
	);
}

/++
 + 平行移動行列を作成
 +
 + Params:
 +  x = X軸移動量
 +  y = Y軸移動量
 +  z = Z軸移動量
 +
 + Returns:
 +  平行移動行列
 +/
Matrix createTranslationMatrix(float x, float y, float z)
{
	return Matrix
	(
		1, 0, 0, 0,
		0, 1, 0, 0,
		0, 0, 1, 0,
		x, y, z, 1
	);
}

/++
 + X軸回転行列を作成
 +
 + Params:
 +  radian = 角度
 +
 + Returns:
 +  X軸回転行列
 +/
Matrix createXRotationMatrix(float radian)
{
	float sin = std.math.sin(radian);
	float cos = std.math.cos(radian);

	return Matrix
	(
		1, 0, 0, 0,
		0, cos, sin, 0,
		0, -sin, cos, 0,
		0, 0, 0, 1
	);
}

/++
 + Y軸回転行列を作成
 +
 + Params:
 +  radian = 角度
 +
 + Returns:
 +  Y軸回転行列
 +/
Matrix createYRotationMatrix(float radian)
{
	float sin = std.math.sin(radian);
	float cos = std.math.cos(radian);

	return Matrix
	(
		cos, 0, -sin, 0,
		0, 1, 0, 0,
		sin, 0, cos, 0,
		0, 0, 0, 1
	);
}

/++
 + Z軸回転行列を作成
 +
 + Params:
 +  radian = 角度
 +
 + Returns:
 +  Z軸回転行列
 +/
Matrix createZRotationMatrix(float radian)
{
	float sin = std.math.sin(radian);
	float cos = std.math.cos(radian);

	return Matrix
	(
		cos, sin, 0, 0,
		-sin, cos, 0, 0,
		0, 0, 1, 0,
		0, 0, 0, 1
	);
}

